#!/usr/bin/env python3

# Copyright (C) 2022-2023:
#   Helmholtz-Zentrum Potsdam Deutsches GeoForschungsZentrum GFZ
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import abc
import logging
import sqlite3
import psycopg2
from packaging import version
from uuid import uuid4

from quadkeyfunctions import get_quadkey_functions

logger = logging.getLogger()


class AbstractDatabase(abc.ABC):
    """
    The AbstractDatabase class represents any database. It manages the
    database connection and cursor.

    Attributes:
        connection:
            Database connection.
        cursor:
            Database cursor.
    """

    def __init__(self):
        self.connection = None
        self.cursor = None

    @abc.abstractmethod
    def connect(self):
        """
        Abstract connection to the database.
        """

        pass

    def commit_and_close(self):
        """
        Commits the executed statements and closes the database connection.
        """

        self.connection.commit()
        self.close()

    def close(self):
        """
        Closes the database connection.
        """

        self.connection.close()


class SpatiaLiteDatabase(AbstractDatabase):
    """
    The `SpatiaLiteDatabase` class represents a SpatiaLite database. It manages the
    database connection and cursor. Upon connection, the SpatiaLite extension is loaded.

    Args:
        database_filepath (str):
            Filepath for the SpatiaLite database file.
        spatialite_filepath (str, default: 'mod_spatialite'):
            Filepath of the SpatiaLite extension.

    Attributes:
        database_filepath (str):
            SpatiaLite database filepath.
        spatialite_filepath (str):
            Filepath to the SpatiaLite extension.
        connection (sqlite3.Connection):
            Database connection
        cursor (sqlite3.Cursor):
            Database cursor
    """

    def __init__(self, database_filepath, spatialite_filepath="mod_spatialite"):
        super().__init__()
        self.database_filepath = database_filepath
        self.spatialite_filepath = spatialite_filepath

    def connect(self, init_spatial_metadata=True, **kwargs):
        """
        Connects to a database, loads the SpatiaLite extension and
        initializes it if requested. If the database file does not exist,
        it will be automatically created.

        Args:
            init_spatial_metadata (Bool):
                Defines if the spatial metadata should be initialized. This is not
                necessary if an existing SpatiaLite database is opened.
        """

        # Connect (if exists) or create SQLite database
        logger.debug("Connecting to database at %s" % self.database_filepath)
        try:
            self.connection = sqlite3.connect(self.database_filepath, **kwargs)
        except Exception:
            logger.exception("Error connecting to the database file")
            raise
        logger.debug("Connection to database established")

        # Load and initialize the SpatiaLite extension
        self.connection.enable_load_extension(True)
        sql_statement = "SELECT load_extension('%s');" % self.spatialite_filepath
        try:
            self.connection.execute(sql_statement)
        except sqlite3.OperationalError:
            logger.exception("Error loading spatial extension")
            raise
        logger.debug("SpatiaLite extension loaded")

        if init_spatial_metadata:
            try:
                self.connection.execute("SELECT InitSpatialMetaData(1);")
            except Exception:
                logger.exception("Error initializing spatial metadata")
                raise
            logger.debug("Spatial metadata initialized")

        # Debug output
        try:
            versions = self.connection.execute("SELECT sqlite_version(), spatialite_version()")
        except Exception:
            logger.exception("Error querying database versions")
            raise
        for row in versions:
            logger.debug("SQLite version: %s" % row[0])
            logger.debug("SpatiaLite version: %s" % row[1])

        try:
            self.connection.execute("PRAGMA foreign_keys = ON")
        except Exception:
            logger.exception("Error with enabling foreign keys")
            raise

        self.cursor = self.connection.cursor()

    @staticmethod
    def check_sqlite_support(requested_version):
        """
        Check if the local SQLite version is higher or equal than the input `version`.

        Args:
            requested_version (str):
                Requested version of SQLite, e.g. `3.35`

        Returns:
            `True` if the local SQLite version is higher than or the same as the requested
            version, otherwise `False`.
        """

        return version.parse(sqlite3.sqlite_version) >= version.parse(requested_version)

    def __copy__(self):
        return SpatiaLiteDatabase(self.database_filepath, self.spatialite_filepath)


class PostGISDatabase(AbstractDatabase):
    """
    The PostGISDatabase class represents a PostGIS database. It manages the
    database connection and cursor.

    Args:
        host (str):
            Host of the database.
        dbname (str):
            Name of the database.
        port (str):
            Port of the database.
        username (str):
            Username for the database connection.
        password (str):
            Password for the database connection.
        timeout (int, optional, default: 0):
            Timeout for SQL queries in milliseconds (`0` equals to no timeout).
        itersize (int, optional, default: 10000):
            Number of rows fetched from the server at one call.

    Attributes:
        host (str):
            Host of the database.
        dbname (str):
            Name of the database.
        port (str):
            Port of the database.
        username (str):
            Username for the database connection.
        password (str):
            Password for the database connection.
        timeout (int):
            Default timeout for SQL queries in milliseconds.
        itersize (int):
            Number of rows fetched from the server at one call.
        connection (psycopg2.Connection):
            Database connection.
        cursor (psycopg2.Cursor):
            Database cursor.
    """

    def __init__(
        self, host, dbname, port, username, password, timeout: int = 0, itersize: int = 10000
    ):
        super().__init__()
        self.host = host
        self.dbname = dbname
        self.port = port
        self.username = username
        self.password = password
        self.timeout = int(timeout)
        self.itersize = int(itersize)

    def connect(self, timeout: int = None):
        """
        Creates a database connection and a cursor with psycopg2 with the
        given credentials and include the connection and cursor as new attributes.

        Args:
            timeout (int, optional):
                Timeout value for SQL queries in milliseconds.
        """

        connection_string = (
            f"host={self.host} "
            f"dbname={self.dbname} "
            f"user={self.username} "
            f"password={self.password} "
            f"port={self.port}"
        )
        try:
            if timeout:
                self.connection = psycopg2.connect(
                    connection_string, options=f"-c statement_timeout={timeout * 1000}"
                )
            elif self.timeout:
                self.connection = psycopg2.connect(
                    connection_string, options=f"-c statement_timeout={self.timeout * 1000}"
                )
            else:
                self.connection = psycopg2.connect(connection_string)
            self.cursor = self.connection.cursor()
        except Exception as e:
            logger.exception(e)

    def execute_server_side(self, sql_statement: str, itersize: int = None):
        """
        Execute an SQL statement with a server-side cursor.

        Args:
            sql_statement (str):
                SQL statement to be executed.
            itersize (int):
                Number of rows fetched from the server at one call.

        Returns:
            Generator object of the query result.
        """

        with self.connection.cursor(name=f"{uuid4()}") as cursor:
            if itersize is not None:
                cursor.itersize = itersize
            else:
                cursor.itersize = self.itersize
            cursor.execute(sql_statement)
            for result in cursor:
                yield result

    def implement_quadkey_functions(self):
        """
        Get all Quadkey functions from the `quadkeyfunctions` module and implement them in the
        PostGIS database.
        """

        if self.connection is None:
            raise AssertionError("PostGIS database not connected.")

        for sql_function in get_quadkey_functions():
            logger.info(f"Implementing function:\n{sql_function}")
            self.cursor.execute(sql_function)

    def __copy__(self):
        return PostGISDatabase(
            self.host,
            self.dbname,
            self.port,
            self.username,
            self.password,
            self.timeout,
            self.itersize,
        )
