-- Copyright (C) 2023:
--   Helmholtz-Zentrum Potsdam Deutsches GeoForschungsZentrum GFZ
--
-- This program is free software: you can redistribute it and/or modify it
-- under the terms of the GNU Affero General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or (at
-- your option) any later version.
--
-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
-- General Public License for more details.
--
-- You should have received a copy of the GNU Affero General Public License
-- along with this program. If not, see http://www.gnu.org/licenses/.

-- Retrieve all Quadkey tiles that intersect with the bounding box of a geometry.
-- Adapted from https://learn.microsoft.com/en-us/bingmaps/articles/bing-maps-tile-system

CREATE OR REPLACE FUNCTION geom_to_quadkeys(
    IN geom GEOMETRY,
    IN level_of_detail INT
)
RETURNS TABLE (quadkey TEXT)
LANGUAGE plpgsql IMMUTABLE PARALLEL SAFE
AS $$
DECLARE
    _bbox GEOMETRY;
    _min_lat DOUBLE PRECISION;
    _min_lon DOUBLE PRECISION;
    _max_lat DOUBLE PRECISION;
    _max_lon DOUBLE PRECISION;
    _min_pixel_xy RECORD;
    _max_pixel_xy RECORD;
    _min_tile_xy RECORD;
    _max_tile_xy RECORD;
BEGIN
    _bbox = ST_Envelope(geom);
    _min_lat = ST_YMin(_bbox);
    _max_lat = ST_YMax(_bbox);
    _min_lon = ST_XMin(_bbox);
    _max_lon = ST_XMax(_bbox);
    _min_pixel_xy = latlon_to_pixel_xy(_max_lat, _min_lon, level_of_detail);
    _max_pixel_xy = latlon_to_pixel_xy(_min_lat, _max_lon, level_of_detail);
    _min_tile_xy = pixel_xy_to_tile_xy(_min_pixel_xy.pixel_x, _min_pixel_xy.pixel_y);
    _max_tile_xy = pixel_xy_to_tile_xy(_max_pixel_xy.pixel_x, _max_pixel_xy.pixel_y);
    FOR i IN _min_tile_xy.tile_x.._max_tile_xy.tile_x LOOP
        FOR j IN _min_tile_xy.tile_y.._max_tile_xy.tile_y LOOP
            quadkey := tile_xy_to_quadkey(i, j, level_of_detail);
            RETURN next;
        END LOOP;
    END LOOP;
END;
$$;
